#include "book_accounting.h"


bool str_Date::operator ==(const str_Date& p_d) const
{
	if (this == &p_d)
		return true;
	return f_year == p_d.f_year
		&& f_month == p_d.f_month
		&& f_day == p_d.f_day;
}

bool str_Date::operator <(const str_Date& p_to_comp) const
{
	if (f_year < p_to_comp.f_year)
	{
		return true;
	}
	else if (f_year == p_to_comp.f_year)
	{
		if (f_month < p_to_comp.f_month)
		{
			return true;
		}
		else if (f_month == p_to_comp.f_month)
		{
			if (f_day < p_to_comp.f_day)
			{
				return true;
			}
			else
			{
				return false;
			}
		}
		else
		{
			return false;
		}
	}
	else
	{
		return false;
	}
}

std::ostream& operator <<(std::ostream& out, const str_Date& p_date)
{
	out << p_date.f_day << "/" << p_date.f_month << "/" << p_date.f_year;
	return out;
}

std::istream& operator >>(std::istream& in, str_Date& p_date)
{
	std::cout << "f_day = ";
	in >> p_date.f_day;
	std::cout << "f_mon = ";
	in >> p_date.f_month;
	std::cout << "f_yr = ";
	in >> p_date.f_year;
	return in;
}

str_Date convert_ctime_to_str_Date(const tm* p_tm)
{
	str_Date result;

	result.f_year = p_tm->tm_year + 1900;
	result.f_month = p_tm->tm_mon + 1;
	result.f_day = p_tm->tm_mday;

	return result;
}

//--------end---------

//--------str_BookAccount--------

struct BookAccountLessFunctor
{
	bool operator()(const str_BookAccount & p_fst, const str_BookAccount & p_scd) const
	{
		if (p_fst.f_library_card_number != p_scd.f_library_card_number)
			return p_fst.f_library_card_number < p_scd.f_library_card_number;
		else if (p_fst.f_book_name != p_scd.f_book_name)
			return p_fst.f_book_name < p_scd.f_book_name;
		else
			return p_fst.f_publishing_house < p_scd.f_publishing_house;
	}
};

std::ostream& operator <<(std::ostream& out, const str_BookAccount& p_date)
{
	out << "Library card number : " << p_date.f_library_card_number << std::endl <<
		"Reader family name  : " << p_date.f_reader_family_name << std::endl <<
		"Date when given     : " << p_date.f_date_giveaway << std::endl <<
		"Days to until return: " << p_date.f_date_return << std::endl <<
		"Book author         : " << p_date.f_book_author << std::endl <<
		"Book name           : " << p_date.f_book_name << std::endl <<
		"Release year        : " << p_date.f_book_year << std::endl <<
		"Publishing house    : " << p_date.f_publishing_house << std::endl <<
		"Price: " << p_date.f_price;

	return out;
}

std::istream& operator >>(std::istream& in, str_BookAccount& p_ba)
{
	std::string tech;
	std::cout << "f_lcn = ";
	in >> p_ba.f_library_card_number;
	std::cout << "f_rfn = " << std::endl;
	in >> tech;
	strcpy_s(p_ba.f_reader_family_name, tech.c_str());
	std::cout << "f_dga = ";
	in >> p_ba.f_date_giveaway;
	std::cout << "f_dr = ";
	in >> p_ba.f_date_return;
	std::cout << "f_bauth = " << std::endl;
	in >> tech;
	strcpy_s(p_ba.f_book_author, tech.c_str());
	std::cout << "f_bn = " << std::endl;
	in >> tech;
	strcpy_s(p_ba.f_book_name, tech.c_str());
	std::cout << "f_by = ";
	in >> p_ba.f_book_year;
	std::cout << "f_ph = " << std::endl;
	in >> tech;
	strcpy_s(p_ba.f_publishing_house, tech.c_str());
	std::cout << "f_pr = ";
	in >> p_ba.f_price;

	return in;
}

//--------end---------

//--------cl_BookAccounter--------

void cl_BookAccounter::load_in()
{
	f_book_accounts_map.clear();
	f_amount = 0;
	std::ifstream in(f_account_base_name + ".bin", std::ios::binary);
	if (!in.is_open())
		return;

	str_BookAccount currentBA;
	while (in.read(reinterpret_cast<char*>(&currentBA), sizeof(str_BookAccount)))
	{
		f_amount++;
		f_book_accounts_map.insert({ f_amount, currentBA });
	}

	in.close();
	

}

bool cl_BookAccounter::load_out()
{
	std::ofstream out(f_account_base_name + ".bin", std::ios::binary);

	std::map<size_t, str_BookAccount>::iterator BAMapIterator = f_book_accounts_map.begin();

	while (BAMapIterator != f_book_accounts_map.end())
	{
		out.write(reinterpret_cast<char*>(&BAMapIterator->second), sizeof(str_BookAccount));
		BAMapIterator++;
	}

	out.close();

	return true;
}

bool cl_BookAccounter::is_outdated(const std::map<size_t, str_BookAccount>::iterator& p_iter, const str_Date& p_current_date)
{
	return p_iter->second.f_date_return < p_current_date;
}

cl_BookAccounter::cl_BookAccounter(std::string&& p_account_base_name) :
	f_account_base_name(p_account_base_name)
{
	load_in();
}

cl_BookAccounter::cl_BookAccounter(const std::string& p_account_base_name) :
	f_account_base_name(p_account_base_name)
{
	load_in();
}

cl_BookAccounter::~cl_BookAccounter()
{
	load_out();

	f_book_accounts_map.clear();
}

bool cl_BookAccounter::is_empty()
{
	return f_book_accounts_map.size() == 0;
}

bool cl_BookAccounter::add_account(const str_BookAccount& p_b_acc)
{
	f_amount++;
	std::pair<size_t, str_BookAccount> c_p(f_amount, p_b_acc);
	f_book_accounts_map.insert(c_p);
	return true;
}

bool cl_BookAccounter::add_account(str_BookAccount&& p_b_acc)
{
	f_amount++;
	f_book_accounts_map.emplace(std::make_pair(f_amount, p_b_acc));
	return true;
}

bool cl_BookAccounter::search_account(const str_BookAccount& p_b_acc)
{
	auto res = std::find_if(f_book_accounts_map.begin(), f_book_accounts_map.end(), [&p_b_acc](const std::pair<size_t, str_BookAccount>& p_BA)
		{
			return p_BA.second == p_b_acc;
		});
	if (res == f_book_accounts_map.end())
	{
		return false;
	}

	std::ofstream res_out(f_account_base_name + "LastOperation.bin", std::ios::binary | std::ios::trunc);

	while (res != f_book_accounts_map.end())
	{
		res_out.write(reinterpret_cast<char*>(&res->second), sizeof(str_BookAccount));
		std::cout << res->second << std::endl << std::endl;
		res++;
		res = std::find_if(res, f_book_accounts_map.end(), [&p_b_acc](const std::pair<size_t, str_BookAccount>& p_BA)
			{
				return p_BA.second == p_b_acc;
			});
	}

	res_out.close();

	return true;
}

bool cl_BookAccounter::search_account(const uint16_t& p_library_card_number)
{
	auto res = std::find_if(f_book_accounts_map.begin(), f_book_accounts_map.end(), [&p_library_card_number](const std::pair<size_t, str_BookAccount>& p_BA)
		{
			return p_BA.second.f_library_card_number == p_library_card_number;
		});
	if (res == f_book_accounts_map.end())
	{
		return false;
	}

	std::ofstream res_out(f_account_base_name + "LastOperation.bin", std::ios::binary | std::ios::trunc);

	while (res != f_book_accounts_map.end())
	{
		res_out.write(reinterpret_cast<char*>(& res->second), sizeof(str_BookAccount));
		std::cout << res->second << std::endl << std::endl;
		res++;
		res = std::find_if(res, f_book_accounts_map.end(), [&p_library_card_number](const std::pair<size_t, str_BookAccount>& p_BA)
			{
				return p_BA.second.f_library_card_number == p_library_card_number;
			});
	}

	res_out.close();

	return true;
}

bool cl_BookAccounter::search_account_auth(const std::string& p_book_author)
{
	auto res = std::find_if(f_book_accounts_map.begin(), f_book_accounts_map.end(), [&p_book_author](const std::pair<size_t, str_BookAccount>& p_BA)
		{
			return p_BA.second.f_book_author == p_book_author;
		});
	if (res == f_book_accounts_map.end())
	{
		return false;
	}

	std::ofstream res_out(f_account_base_name + "LastOperation.bin", std::ios::binary | std::ios::trunc);

	while (res != f_book_accounts_map.end())
	{
		res_out.write(reinterpret_cast<char*>(&res->second), sizeof(str_BookAccount));
		std::cout << res->second << std::endl << std::endl;
		res++;
		res = std::find_if(res, f_book_accounts_map.end(), [&p_book_author](const std::pair<size_t, str_BookAccount>& p_BA)
			{
				return p_BA.second.f_book_author == p_book_author;
			});
	}

	res_out.close();

	return true;
}

bool cl_BookAccounter::search_account_ph(const std::string& p_publishing_house)
{
	auto res = std::find_if(f_book_accounts_map.begin(), f_book_accounts_map.end(), [&p_publishing_house](const std::pair<size_t, str_BookAccount>& p_BA)
		{
			return p_BA.second.f_publishing_house == p_publishing_house;
		});
	if (res == f_book_accounts_map.end())
	{
		return false;
	}

	std::ofstream res_out(f_account_base_name + "LastOperation.bin", std::ios::binary | std::ios::trunc);

	while (res != f_book_accounts_map.end())
	{
		res_out.write(reinterpret_cast<char*>(&res->second), sizeof(str_BookAccount));
		std::cout << res->second << std::endl << std::endl;
		res++;
		res = std::find_if(res, f_book_accounts_map.end(), [&p_publishing_house](const std::pair<size_t, str_BookAccount>& p_BA)
			{
				return p_BA.second.f_publishing_house == p_publishing_house;
			});
	}

	res_out.close();

	return true;
}

void cl_BookAccounter::peek_last_result()
{
	std::ifstream res_in(f_account_base_name + "LastOperation.bin", std::ios::binary);

	str_BookAccount currentBA;

	while (res_in.read(reinterpret_cast<char*>(&currentBA), sizeof(str_BookAccount)))
	{
		std::cout << currentBA << std::endl << std::endl;
	}

	res_in.close();
}

unsigned int cl_BookAccounter::check_outdated_accounts(const str_Date& p_current_date, bool is_with_printout)
{
	std::ofstream res_out(f_account_base_name + "LastOperation.bin", std::ios::binary | std::ios::trunc);
	unsigned int count = 0;
	auto res = f_book_accounts_map.begin();
	while (res != f_book_accounts_map.end())
	{
		if (is_outdated(res, p_current_date))
		{
			res_out.write(reinterpret_cast<char*>(&res->second), sizeof(str_BookAccount));
			if(is_with_printout)
				std::cout << res->second << std::endl << std::endl;

			count++;
		}

		res++;
	}
	res_out.close();

	return count;
}

bool cl_BookAccounter::delete_account(const str_BookAccount& p_b_acc)
{
	auto res = std::find_if(f_book_accounts_map.begin(), f_book_accounts_map.end(), [&p_b_acc](const std::pair<size_t, str_BookAccount>& p_BA)
		{
			return p_BA.second == p_b_acc;
		});
	if (res == f_book_accounts_map.end())
	{
		return false;
	}
	
	while (res != f_book_accounts_map.end())
	{
		res = f_book_accounts_map.erase(res);
		f_amount--;
		res = std::find_if(res, f_book_accounts_map.end(), [&p_b_acc](const std::pair<size_t, str_BookAccount>& p_BA)
			{
				return p_BA.second == p_b_acc;
			});
	}
	return true;
}

bool cl_BookAccounter::delete_account(const unsigned int& p_b_acc_num)
{
	if (p_b_acc_num > f_book_accounts_map.size())
	{
		throw std::exception("Out of bounds");
	}

	auto res = f_book_accounts_map.begin();
	unsigned int counter = p_b_acc_num;
	while (res != f_book_accounts_map.end() && counter > 1)
	{
		res++;
		counter--;
	}

	f_book_accounts_map.erase(res);
	f_amount--;

	return true;
}

bool cl_BookAccounter::change_account(const str_BookAccount& p_b_acc_to_change, const str_BookAccount& p_b_acc_to_imply)
{
	auto res = std::find_if(f_book_accounts_map.begin(), f_book_accounts_map.end(), [&p_b_acc_to_change](const std::pair<size_t, str_BookAccount>& p_BA)
		{
			return p_BA.second == p_b_acc_to_change;
		});
	if (res == f_book_accounts_map.end())
	{
		return false;
	}
	
	while (res != f_book_accounts_map.end())
	{
		res->second = p_b_acc_to_imply;
		res++;
		res = std::find_if(f_book_accounts_map.begin(), f_book_accounts_map.end(), [&p_b_acc_to_change](const std::pair<size_t, str_BookAccount>& p_BA)
			{
				return p_BA.second == p_b_acc_to_change;
			});
	}

	return true;
}

bool cl_BookAccounter::change_account(const str_BookAccount& p_b_acc_to_change, str_BookAccount&& p_b_acc_to_imply)
{
	auto res = std::find_if(f_book_accounts_map.begin(), f_book_accounts_map.end(), [&p_b_acc_to_change](const std::pair<size_t, str_BookAccount>& p_BA)
		{
			return p_BA.second == p_b_acc_to_change;
		});
	if (res == f_book_accounts_map.end())
	{
		return false;
	}

	while (res != f_book_accounts_map.end())
	{
		res->second = p_b_acc_to_imply;
		res++;
		res = std::find_if(f_book_accounts_map.begin(), f_book_accounts_map.end(), [&p_b_acc_to_change](const std::pair<size_t, str_BookAccount>& p_BA)
			{
				return p_BA.second == p_b_acc_to_change;
			});
	}

	return true;
}

void cl_BookAccounter::load_out_txt(const str_Date& p_current_date)
{
	auto iter = f_book_accounts_map.begin();
	std::ofstream txt_out(f_account_base_name + ".txt");

	txt_out << std::setw(159) << std::setfill('-') << "-" << std::endl;
	txt_out << std::setfill(' ');
	txt_out << std::setw(90) << " Accounts " << std::endl;
	txt_out << std::setw(159) << std::setfill('-') << "-" << std::endl;
	txt_out << std::setfill(' ');
	txt_out << "| Library card number | Reader family name  | Date when given | Days to until return | Book author | Book name| Release year  | Publishing house | Price | " << std::endl;
	txt_out << std::setw(159) << std::setfill('-') << "-" << std::endl;
	txt_out << std::setw(1) << std::setfill(' ');


	while (iter != f_book_accounts_map.end())
	{

		txt_out << "|" << std::setw(17) << iter->second.f_library_card_number << " |"
			<< std::setw(14) << iter->second.f_reader_family_name << " |"
			<< std::setw(9) << iter->second.f_date_giveaway << " |"
			<< std::setw(9) << iter->second.f_date_return << " |"
			<< std::setw(22) << iter->second.f_book_author << " |"
			<< std::setw(13) << iter->second.f_book_name << " |"
			<< std::setw(12) << iter->second.f_book_year << " |"
			<< std::setw(14) << iter->second.f_publishing_house << " |"
			<< std::setw(10) << iter->second.f_price << " |"
			<< std::endl;

		iter++;

	}

    //!!! По заданию нужно вывести в виде отформатированной (!) таблицы с шапкой.->Исправлено

}

//--------end---------
