
#include <iostream>
#include <string>
#include <fstream>
#include <iterator>
#include <algorithm>
#include <functional>
#include <iomanip>
#include <math.h>
#include <map>
#include <ctime>

struct str_Date
{
	unsigned int f_year;
	unsigned int f_month;
	unsigned int f_day;

	str_Date() :
		f_day(0),
		f_month(0),
		f_year(0)
	{}

	str_Date(unsigned int p_year, 
		unsigned int p_month,
		unsigned int p_day) :
		f_day(p_day),
		f_month(p_month),
		f_year(p_year)
	{}

	bool operator ==(const str_Date& p_d) const;

	bool operator <(const str_Date& p_to_comp) const;

	friend std::ostream& operator <<(std::ostream& out, const str_Date& p_date);

	friend std::istream& operator >>(std::istream& in, str_Date& p_date);
};

str_Date convert_ctime_to_str_Date(const tm* p_tm);

struct str_BookAccount {

	uint16_t f_library_card_number;
	char f_reader_family_name[30];
	str_Date f_date_giveaway;
	str_Date f_date_return;
	char f_book_author[30];
	char f_book_name[30];
	unsigned int f_book_year;
	char f_publishing_house[30];
	double f_price;


	str_BookAccount() :
		f_book_year(0),
		f_date_giveaway(str_Date()),
		f_date_return(str_Date()),
		f_library_card_number(0),
		f_price(0),
		f_book_author(""),
		f_book_name(""),
		f_publishing_house(""),
		f_reader_family_name("")
	{}

	str_BookAccount(uint16_t p_library_card_number,
		const std::string& p_reader_family_name,
		str_Date p_date_giveaway,
		str_Date p_date_return,
		const std::string& p_book_author,
		const std::string& p_book_name,
		unsigned int p_book_year,
		const std::string& p_publishing_house,
		double p_price) :
		f_book_year(p_book_year),
		f_date_giveaway(p_date_giveaway),
		f_date_return(p_date_return),
		f_library_card_number(p_library_card_number),
		f_price(p_price)
	{
		strcpy_s(f_book_author, 29, p_book_author.c_str());
		strcpy_s(f_book_name, 29, p_book_name.c_str());
		strcpy_s(f_publishing_house, 29, p_publishing_house.c_str());
		strcpy_s(f_reader_family_name, 29, p_reader_family_name.c_str());
	}

	str_BookAccount(uint16_t&& p_library_card_number,
		const char* p_reader_family_name,
		str_Date&& p_date_giveaway,
		str_Date&& p_date_return,
		const char* p_book_author,
		const char* p_book_name,
		unsigned int&& p_book_year,
		const char* p_publishing_house,
		double&& p_price) :
		f_book_year(std::move(p_book_year)),
		f_date_giveaway(std::move(p_date_giveaway)), //!!! Если вы используете r-value ссылки, то где move?->Исправлено
		f_date_return(std::move(p_date_return)),      //!!! Если вы используете r-value ссылки, то где move?->Исправлено
		f_library_card_number(std::move(p_library_card_number)),
		f_price(std::move(p_price))
	{
		strcpy_s(f_book_author, 29, p_book_author);
		strcpy_s(f_book_name, 29, p_book_name);
		strcpy_s(f_publishing_house, 29, p_publishing_house);
		strcpy_s(f_reader_family_name, 29, p_reader_family_name);
	}

	str_BookAccount& operator = (const str_BookAccount& p_to_f) 
	{
		strcpy_s(f_book_author, p_to_f.f_book_author);
		strcpy_s(f_book_name, p_to_f.f_book_name);
		f_book_year = p_to_f.f_book_year;
		f_date_giveaway = p_to_f.f_date_giveaway;
		f_date_return = p_to_f.f_date_return;
		f_library_card_number = p_to_f.f_library_card_number;
		f_price =p_to_f.f_price;
		strcpy_s(f_publishing_house, p_to_f.f_publishing_house);
		strcpy_s(f_reader_family_name, p_to_f.f_reader_family_name);
		return *this;
	}

	bool operator == (const str_BookAccount& p_to_f) const
	{
		return strcmp(f_book_author, p_to_f.f_book_author) == 0
			&& strcmp(f_book_name, p_to_f.f_book_name) == 0
			&& f_book_year == p_to_f.f_book_year
			&& f_date_giveaway == p_to_f.f_date_giveaway
			&& f_date_return == p_to_f.f_date_return
			&& f_library_card_number == p_to_f.f_library_card_number
			&& f_price == p_to_f.f_price
			&& strcmp(f_publishing_house, p_to_f.f_publishing_house) == 0
			&& strcmp(f_reader_family_name, p_to_f.f_reader_family_name) == 0;
	}
	
	void printout() const
	{
		std::cout << std::endl <<
			"Library card number : " << f_library_card_number << std::endl <<
			"Reader family name  : " << f_reader_family_name << std::endl <<
			"Date when given     : " << f_date_giveaway.f_day << "/" << f_date_giveaway.f_month << "/" << f_date_giveaway.f_year << std::endl <<
			"Days to until return: " << f_date_return.f_day << "/" << f_date_return.f_month << "/" << f_date_return.f_year << std::endl <<
			"Book author         : " << f_book_author << std::endl <<
			"Book name           : " << f_book_name << std::endl <<
			"Release year        : " << f_book_year << std::endl <<
			"Publishing house    : " << f_publishing_house << std::endl <<
			"Price: " << f_price << std::endl << std::endl;
	}

	friend std::ostream& operator <<(std::ostream& out, const str_BookAccount& p_date);

	friend std::istream& operator >>(std::istream& in, str_BookAccount& p_date);
};

struct BookAccountLessFunctor;

class cl_BookAccounter
{
private:

	std::string f_account_base_name;

	std::map<size_t, str_BookAccount> f_book_accounts_map;
	size_t f_amount;

	void load_in();

	bool load_out();

	bool is_outdated(const std::map<size_t, str_BookAccount>::iterator& p_iter, const str_Date& p_current_date);

public:

	cl_BookAccounter(std::string&& p_account_base_name);

	cl_BookAccounter(const std::string& p_account_base_name);

	~cl_BookAccounter();

	bool is_empty();

	bool add_account(const str_BookAccount& p_b_acc);

	bool add_account(str_BookAccount&& p_b_acc);

	bool search_account(const str_BookAccount& p_b_acc);

	bool search_account(const uint16_t& p_library_card_number);

	bool search_account_auth(const std::string& p_book_author);

	bool search_account_ph(const std::string& p_publishing_house);

	void peek_last_result();

	unsigned int check_outdated_accounts(const str_Date& p_current_date, bool is_with_printout);

	bool delete_account(const str_BookAccount& p_b_acc);

	bool delete_account(const unsigned int& p_b_acc_num);

	bool change_account(const str_BookAccount& p_b_acc_to_change, const str_BookAccount& p_b_acc_to_imply);

	bool change_account(const str_BookAccount& p_b_acc_to_change, str_BookAccount&& p_b_acc_to_imply);

	void load_out_txt(const str_Date& p_current_date);

};
